/*
AppleWin : An Apple //e emulator for Windows

Copyright (C) 1994-1996, Michael O'Brien
Copyright (C) 1999-2001, Oliver Schmidt
Copyright (C) 2002-2005, Tom Charlesworth
Copyright (C) 2006-2007, Tom Charlesworth, Michael Pohoreski

AppleWin is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

AppleWin is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with AppleWin; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* Description: Keyboard emulation
 *
 * Author: Various
 */

#include "StdAfx.h"


static bool g_bKeybBufferEnable = false;

#define KEY_OLD

static BYTE asciicode[2][10] = {
	{0x08,0x0D,0x15,0x2F,0x00,0x00,0x00,0x00,0x00,0x00},
	{0x08,0x0B,0x15,0x0A,0x00,0x00,0x00,0x00,0x00,0x7F}
};	// Convert PC arrow keys to Apple keycodes

bool  g_bShiftKey = false;
bool  g_bCtrlKey  = false;
bool  g_bAltKey   = false;

static bool  g_bCapsLock = true; //Caps lock key for Apple2 and Lat/Cyr lock for Pravets8
static bool  g_bP8CapsLock = true; //Caps lock key of Pravets 8A/C
static int   lastvirtkey     = 0;	// Current PC keycode
static BYTE  keycode         = 0;	// Current Apple keycode
static DWORD keyboardqueries = 0;

#ifdef KEY_OLD
// Original
static BOOL  keywaiting      = 0;
#else
// Buffered key input:
// - Needed on faster PCs where aliasing occurs during short/fast bursts of 6502 code.
// - Keyboard only sampled during 6502 execution, so if it's run too fast then key presses will be missed.
const int KEY_BUFFER_MIN_SIZE = 1;
const int KEY_BUFFER_MAX_SIZE = 2;
static int g_nKeyBufferSize = KEY_BUFFER_MAX_SIZE;	// Circ key buffer size
static int g_nNextInIdx = 0;
static int g_nNextOutIdx = 0;
static int g_nKeyBufferCnt = 0;

static struct
{
	int nVirtKey;
	BYTE nAppleKey;
} g_nKeyBuffer[KEY_BUFFER_MAX_SIZE];
#endif

static BYTE g_nLastKey = 0x00;
#if 1	// +CU version
static DWORD g_nArrowKeys = 0;
static char g_szArrowKeysFilename[MAX_PATH] = { 0 };

static struct arrowkeys_file_list_t default_arrowkeys_file_list1[4] = {
	{
		&(default_arrowkeys_file_list1[1]),
		"aklabeth",
		false
	},
	{
		&(default_arrowkeys_file_list1[2]),
		"u1Mater",
		false
	},
	{
		&(default_arrowkeys_file_list1[3]),
		"U2Prog",
		false
	},
	{
		NULL,
		"u3master",
		false
	}
};

static struct arrowkeys_file_list_t default_arrowkeys_file_list2[5] = {
	{
		&(default_arrowkeys_file_list2[1]),
		"AP_WIZ1",
		true
	},
	{
		&(default_arrowkeys_file_list2[2]),
		"AP_WIZ2",
		true
	},
	{
		&(default_arrowkeys_file_list2[3]),
		"AP_WIZ3",
		true
	},
	{
		&(default_arrowkeys_file_list2[4]),
		"AP_WIZ4",
		true
	},
	{
		NULL,
		"AP_WIZ5",
		true
	}
};

static struct arrowkeys_list_t default_arrowkeys_list[9] = {
	{
		&(default_arrowkeys_list[1]),
		default_arrowkeys_file_list1,
		"up=RETURN down=/",
		{
			{ VK_LEFT	, NOT_ASCII	},
			{ 0x0D		, ASCII		},
			{ VK_RIGHT	, NOT_ASCII	},
			{ '/'		, ASCII		}
		}
	},
	{
		&(default_arrowkeys_list[2]),
		NULL,
		"up=[ left=; right=' down=/",
		{
			{ ';'		, ASCII		},
			{ '['		, ASCII		},
			{ '\''		, ASCII		},
			{ '/'		, ASCII		}
		}
	},
	{
		&(default_arrowkeys_list[3]),
		default_arrowkeys_file_list2,
		"up=W left=A right=D down=X",
		{
			{ 'A'		, ASCII		},
			{ 'W'		, ASCII		},
			{ 'D'		, ASCII		},
			{ 'X'		, ASCII		}
		}
	},
	{
		&(default_arrowkeys_list[4]),
		NULL,
		"up=3 left=W right=E down=S",
		{
			{ 'W'		, ASCII		},
			{ '3'		, ASCII		},
			{ 'E'		, ASCII		},
			{ 'S'		, ASCII		}
		}
	},
	{
		&(default_arrowkeys_list[5]),
		NULL,
		"up=N left=W right=E down=S",
		{
			{ 'W'		, ASCII		},
			{ 'N'		, ASCII		},
			{ 'E'		, ASCII		},
			{ 'S'		, ASCII		}
		}
	},
	{
		&(default_arrowkeys_list[6]),
		NULL,
		"up=F left=L right=R down=B",
		{
			{ 'L'		, ASCII		},
			{ 'F'		, ASCII		},
			{ 'R'		, ASCII		},
			{ 'B'		, ASCII		}
		}
	},
	{
		&(default_arrowkeys_list[7]),
		NULL,
		"up=A down=Z",
		{
			{ VK_LEFT	, NOT_ASCII	},
			{ 'A'		, ASCII		},
			{ VK_RIGHT	, NOT_ASCII	},
			{ 'Z'		, ASCII		}
		}
	},
	{
		&(default_arrowkeys_list[8]),
		NULL,
		"up=I left=J right=L down=K",
		{
			{ 'J'		, ASCII		},
			{ 'I'		, ASCII		},
			{ 'L'		, ASCII		},
			{ 'K'		, ASCII		}
		}
	},
	{
		NULL,
		NULL,
		"up=I left=J right=K down=M",
		{
			{ 'J'		, ASCII		},
			{ 'I'		, ASCII		},
			{ 'K'		, ASCII		},
			{ 'M'		, ASCII		}
		}
	}
};

static const int default_arrowkeys_list_size = 9;

static struct arrowkeys_list_t *g_arrowkeys_list = NULL;
static int g_arrowkeys_list_size = 0;

static const struct arrowkeys_replacekey_t default_replacekeys[4] = {
	{ VK_LEFT	, NOT_ASCII	},
	{ VK_UP		, NOT_ASCII	},
	{ VK_RIGHT	, NOT_ASCII	},
	{ VK_DOWN	, NOT_ASCII	}
};

static const struct arrowkeys_replacekey_t *current_replacekeys = default_replacekeys;

#endif	// +CU version

//
// ----- ALL GLOBALLY ACCESSIBLE FUNCTIONS ARE BELOW THIS LINE -----
//

//===========================================================================

void KeybReset()
{
#ifdef KEY_OLD
	keywaiting = 0;
#else
	g_nNextInIdx = 0;
	g_nNextOutIdx = 0;
	g_nKeyBufferCnt = 0;
	g_nLastKey = 0x00;

	g_nKeyBufferSize = g_bKeybBufferEnable ? KEY_BUFFER_MAX_SIZE : KEY_BUFFER_MIN_SIZE;
#endif
}

//===========================================================================

//void KeybSetBufferMode(bool bNewKeybBufferEnable)
//{
//	if(g_bKeybBufferEnable == bNewKeybBufferEnable)
//		return;
//
//	g_bKeybBufferEnable = bNewKeybBufferEnable;
//	KeybReset();
//}
//
//bool KeybGetBufferMode()
//{
//	return g_bKeybBufferEnable;
//}

//===========================================================================
bool KeybGetAltStatus ()
{
	return g_bAltKey;
}

//===========================================================================
bool KeybGetCapsStatus ()
{
	return g_bCapsLock;
}
//===========================================================================
bool KeybGetP8CapsStatus()
{
	return g_bP8CapsLock;
}
//===========================================================================
/*
bool KeybGetCapsAllowed() //For Pravets 8A/C only
{
	return g_CapsLockAllowed;
}
*/
//===========================================================================
bool KeybGetCtrlStatus ()
{
	return g_bCtrlKey;
}

//===========================================================================
bool KeybGetShiftStatus ()
{
	return g_bShiftKey;
}

//===========================================================================
void KeybUpdateCtrlShiftStatus()
{
	g_bShiftKey = (GetKeyState( VK_SHIFT  ) & KF_UP) ? true : false; // 0x8000 KF_UP
	g_bCtrlKey  = (GetKeyState( VK_CONTROL) & KF_UP) ? true : false;
	g_bAltKey   = (GetKeyState( VK_MENU   ) & KF_UP) ? true : false;
}

//===========================================================================
BYTE KeybGetKeycode ()		// Used by MemCheckPaging() & VideoCheckMode()
{
	return keycode;
}

//===========================================================================
DWORD KeybGetNumQueries ()	// Used in determining 'idleness' of Apple system
{
	DWORD result = keyboardqueries;
	keyboardqueries = 0;
	return result;
}

//===========================================================================
void KeybQueueKeypress (int key, BOOL bASCII)
{
#if 1	// +CU version
	if (!bASCII && VK_LEFT <= key && key <= VK_DOWN && g_nArrowKeys != 0)
	{
		bASCII = current_replacekeys[key - VK_LEFT].is_ascii;
		key    = current_replacekeys[key - VK_LEFT].key;
	}
#endif	// +CU version
	if (bASCII == ASCII)
	{
		if (g_bFreshReset && key == VK_CANCEL) // OLD HACK: 0x03
		{
			g_bFreshReset = false;
			return; // Swallow spurious CTRL-C caused by CTRL-BREAK
		}

		g_bFreshReset = false;
		if (key > 0x7F)
			return;

		if (!IS_APPLE2) 
		{
			P8Shift = false;
			if (g_bCapsLock && (key >= 'a') && (key <='z'))
			{
				P8Shift = true;
				keycode = key - 32;
			}
			else
			{
				keycode = key;
			}			

			//The latter line should be applied for Pravtes 8A/C only, but not for Pravets 82/M !!!			
			if ((g_bCapsLock == false) && (key >= 'A') && (key <='Z'))
			{
				P8Shift = true;
				if (g_Apple2Type == A2TYPE_PRAVETS8A) 
					keycode = key + 32;
			}

			//Remap some keys for Pravets82/M
			if (g_Apple2Type == A2TYPE_PRAVETS82)
			{
				if (key == 64) 
					keycode = 96;
				if (key == '^') 
					keycode = '~';

				if (g_bCapsLock == false) //cyrillic letters
				{
					if (key == '`') keycode = '^';
					if (key == 92) keycode = '@'; // \ to @	
					if (key == 124) keycode = 92;
				}
				else //(g_bCapsLock == true) //latin letters
				{
					if (key == 91) keycode = 123;
					if (key == 93) keycode = 125;
					if (key == 124) keycode = 92;
				}
			}
			if (g_Apple2Type == A2TYPE_PRAVETS8M)  //Pravets 8M charset is still uncertain
			{
				if (g_bCapsLock == false) //cyrillic letters
				{
					if (key == '[') keycode = '{';
					if (key == ']') keycode = '}';
					if (key == '`') keycode = '~'; //96= key `~
					if (key == 92) keycode = 96;
				}
				else //latin letters
				{
					if (key == '`') 
						keycode = '^'; //96= key `~
				}
			}
			//Remap some keys for Pravets8A/C, which has a different charset for Pravtes82/M, whose keys MUST NOT be remapped.
			if (g_Apple2Type == A2TYPE_PRAVETS8A) //&& (g_bCapsLock == false))
			{
				if (g_bCapsLock == false) //i.e. cyrillic letters
			    {
					if (key == '[') keycode = '{';
					if (key == ']') keycode = '}';
					if (key == '`') keycode = '~';
					if (key == 92) keycode = 96;
					if (GetCapsLockAllowed() == true)
					{
						if ((key == 92) || (key == 124)) keycode = 96; // to 
						//This shall be rewriten, so that enabling CAPS_LOCK (i.e. F10) will not invert these keys values)
						//The same for latin letters.
						if ((key == '{') || (key == '}') || (key == '~') || (key == 124) || (key == '^') ||  (key == 95))
							P8Shift = true;					
					}
				}
				else //i.e. latin letters
				{
					if (GetCapsLockAllowed()  == false)
					{
						if (key == '{') keycode = '[';
						if (key == '}') keycode = ']';
						if (key == 124) 
							keycode = 92;
						/*if (key == 92) 
							keycode = 124;*/
					//Characters ` and ~ cannot be generated in 7bit character mode, so they are replaced with
					}
					else
					{
						if (key == '{') keycode = 91;
						if (key == '}')	keycode = 93;
						if (key == 124)	keycode = 92;					
						if ((key == '[') || (key == ']') || (key == 92) || (key == '^') || (key == 95))
							P8Shift= true; 
						if (key == 96)	 //This line shall generate sth. else i.e. ` In fact. this character is not generateable by the pravets keyboard.
						{
							keycode = '^';
							P8Shift= true;
						}
						if (key == 126)	keycode = '^';
					}
				}
			}
		}
		else
		{
			if (g_Apple2Type == A2TYPE_PRAVETS8A)
			{
			}
			else
			{
				if (key >= '`')
					keycode = key - 32;
				else
					keycode = key;
			}
		}
		lastvirtkey = LOBYTE(VkKeyScan(key));
	} 
	else //(bASCII != ASCII)
	{
		// Note: VK_CANCEL is Control-Break
		if ((key == VK_CANCEL) && (GetKeyState(VK_CONTROL) < 0))
		{
#ifndef KEY_OLD
			g_nNextInIdx = g_nNextOutIdx = g_nKeyBufferCnt = 0;
#endif
			g_bFreshReset = true;
			CtrlReset();
			return;
		}

		if ((key == VK_INSERT) && (GetKeyState(VK_SHIFT) < 0))
		{
			// Shift+Insert
			ClipboardInitiatePaste();
			return;
		}

		if (!((key >= VK_LEFT) && (key <= VK_DELETE) && asciicode[IS_APPLE2 ? 0 : 1][key - VK_LEFT]))
			return;

		keycode = asciicode[IS_APPLE2 ? 0 : 1][key - VK_LEFT];		// Convert to Apple arrow keycode
		lastvirtkey = key;
	}
#ifdef KEY_OLD
	keywaiting = 1;
#else
	bool bOverflow = false;

	if(g_nKeyBufferCnt < g_nKeyBufferSize)
		g_nKeyBufferCnt++;
	else
		bOverflow = true;

	g_nKeyBuffer[g_nNextInIdx].nVirtKey = lastvirtkey;
	g_nKeyBuffer[g_nNextInIdx].nAppleKey = keycode;
	g_nNextInIdx = (g_nNextInIdx + 1) % g_nKeyBufferSize;

	if(bOverflow)
		g_nNextOutIdx = (g_nNextOutIdx + 1) % g_nKeyBufferSize;
#endif
}

//===========================================================================

static HGLOBAL hglb = NULL;
static LPTSTR lptstr = NULL;
static bool g_bPasteFromClipboard = false;
static bool g_bClipboardActive = false;

void ClipboardInitiatePaste()
{
	if (g_bClipboardActive)
		return;

	g_bPasteFromClipboard = true;
}

static void ClipboardDone()
{
	if (g_bClipboardActive)
	{
		g_bClipboardActive = false;
		GlobalUnlock(hglb);
		CloseClipboard();
	}
}

static void ClipboardInit()
{
	ClipboardDone();

	if (!IsClipboardFormatAvailable(CF_TEXT))
		return;
	
	if (!OpenClipboard(g_hFrameWindow))
		return;
	
	hglb = GetClipboardData(CF_TEXT);
	if (hglb == NULL)
	{	
		CloseClipboard();
		return;
	}

	lptstr = (char*) GlobalLock(hglb);
	if (lptstr == NULL)
	{	
		CloseClipboard();
		return;
	}

	g_bPasteFromClipboard = false;
	g_bClipboardActive = true;
}

static char ClipboardCurrChar(bool bIncPtr)
{
	char nKey;
	int nInc = 1;

	if((lptstr[0] == 0x0D) && (lptstr[1] == 0x0A))
	{
		nKey = 0x0D;
		nInc = 2;
	}
	else
	{
		nKey = lptstr[0];
	}

	if(bIncPtr)
		lptstr += nInc;

	return nKey;
}

//===========================================================================

BYTE __stdcall KeybReadData (WORD, WORD, BYTE, BYTE, ULONG)
{
	keyboardqueries++;

	//

	if(g_bPasteFromClipboard)
		ClipboardInit();

	if(g_bClipboardActive)
	{
		if(*lptstr == 0)
			ClipboardDone();
		else
			return 0x80 | ClipboardCurrChar(false);
	}

	//

#ifdef KEY_OLD
	return keycode | (keywaiting ? 0x80 : 0);
#else
	BYTE nKey = g_nKeyBufferCnt ? 0x80 : 0;
	if(g_nKeyBufferCnt)
	{
		nKey |= g_nKeyBuffer[g_nNextOutIdx].nAppleKey;
		g_nLastKey = g_nKeyBuffer[g_nNextOutIdx].nAppleKey;
	}
	else
	{
		nKey |= g_nLastKey;
	}
	return nKey;
#endif
}

//===========================================================================

BYTE __stdcall KeybReadFlag (WORD, WORD, BYTE, BYTE, ULONG)
{
	keyboardqueries++;

	//

	if(g_bPasteFromClipboard)
		ClipboardInit();

	if(g_bClipboardActive)
	{
		if(*lptstr == 0)
			ClipboardDone();
		else
			return 0x80 | ClipboardCurrChar(true);
	}

	//

#ifdef KEY_OLD
	keywaiting = 0;
	return keycode | ((GetKeyState(lastvirtkey) < 0) ? 0x80 : 0);
#else
	BYTE nKey = (GetKeyState(g_nKeyBuffer[g_nNextOutIdx].nVirtKey) < 0) ? 0x80 : 0;
	nKey |= g_nKeyBuffer[g_nNextOutIdx].nAppleKey;
	if(g_nKeyBufferCnt)
	{
		g_nKeyBufferCnt--;
		g_nNextOutIdx = (g_nNextOutIdx + 1) % g_nKeyBufferSize;
	}
	return nKey;
#endif
}

//===========================================================================
void KeybToggleCapsLock ()
{
	if (!IS_APPLE2)
	{
		g_bCapsLock = (GetKeyState(VK_CAPITAL) & 1);
		FrameRefreshStatus(DRAW_LEDS);
	}
}

//===========================================================================
void KeybToggleP8ACapsLock ()
{
	_ASSERT(g_Apple2Type == A2TYPE_PRAVETS8A);
	P8CAPS_ON = !P8CAPS_ON;
	FrameRefreshStatus(DRAW_LEDS);
	// g_bP8CapsLock= g_bP8CapsLock?false:true; //The same as the upper, but slower
}

//===========================================================================

DWORD KeybGetSnapshot(SS_IO_Keyboard* pSS)
{
	pSS->keyboardqueries	= keyboardqueries;
	pSS->nLastKey			= g_nLastKey;
	return 0;
}

DWORD KeybSetSnapshot(SS_IO_Keyboard* pSS)
{
	keyboardqueries	= pSS->keyboardqueries;
	g_nLastKey		= pSS->nLastKey;
	return 0;
}

#if 1	// +CU version
//===========================================================================
char* KeybGetArrowKeysFilename()
{
	return g_szArrowKeysFilename;
}

//===========================================================================
void KeybSetArrowKeysFilename(const char *filename)
{
	if (*filename) {
		strcpy(g_szArrowKeysFilename, filename);
	} else {
		memset(g_szArrowKeysFilename, 0, sizeof(g_szArrowKeysFilename));
	}
	g_nArrowKeys = 0;
}

//===========================================================================
void KeybFreeArrowKeysList(struct arrowkeys_list_t *list)
{
	if (list != NULL && list != default_arrowkeys_list) {
		struct arrowkeys_list_t *next = list;
		while (next != NULL) {
			struct arrowkeys_list_t *cur = next;
			next = next->next;

			struct arrowkeys_file_list_t *next_file = cur->file;
			while (next_file != NULL) {
				struct arrowkeys_file_list_t *cur_file = next_file;
				next_file = next_file->next;
				if (cur_file->name != NULL) {
					delete[] cur_file->name;
				}
				delete cur_file;
			}

			if (cur->label != NULL) {
				delete[] cur->label;
			}
			delete cur;
		}
	}
}

//===========================================================================
void KeybSetArrowKeysList(struct arrowkeys_list_t *list)
{
	KeybFreeArrowKeysList(g_arrowkeys_list);
	if (list != NULL) {
		g_arrowkeys_list = list;
		for (g_arrowkeys_list_size = 0; list != NULL; list = list->next) {
			g_arrowkeys_list_size++;
		}
	} else {
		g_arrowkeys_list = default_arrowkeys_list;
		g_arrowkeys_list_size = default_arrowkeys_list_size;
	}

	KeybSetArrowKeysIndex(0);
}

//===========================================================================
int KeybGetArrowKeysIndex()
{
	return g_nArrowKeys;
}

//===========================================================================
void KeybSetArrowKeysIndex(int index)
{
	if (0 <= index && index < (g_arrowkeys_list_size + ARROWKEYS_INDEX_OFFSET)) {
		g_nArrowKeys = index;
	} else {
		g_nArrowKeys = 0;
	}
	KeybSelectArrowKeys();
}

//===========================================================================
int KeybLoadArrowKeys()
{
	KeybSetArrowKeysIndex(0);

	if (*g_szArrowKeysFilename != 0) {
		KeybFreeArrowKeysList(g_arrowkeys_list);
		g_arrowkeys_list = NULL;
		g_arrowkeys_list_size = 0;

		if (0 < (g_arrowkeys_list_size = KeybLoadArrowKeys(g_arrowkeys_list, g_szArrowKeysFilename))) {
			return g_arrowkeys_list_size;
		}
	}

	g_arrowkeys_list = default_arrowkeys_list;
	g_arrowkeys_list_size = default_arrowkeys_list_size;

	return 0;
}

//===========================================================================
int KeybLoadArrowKeys(struct arrowkeys_list_t * &list, const char *filename)
{
	int size = 0;

	KeybFreeArrowKeysList(list);
	list = NULL;

	if (filename == NULL || *filename == 0) {
		list = default_arrowkeys_list;
		return default_arrowkeys_list_size;
	}

	FILE *fp = fopen(filename, "rt");
	if (fp == NULL) {
		return -1;
	}

	char buf[1024];
	bool err = false;
	int line = 0;
	struct arrowkeys_list_t *prev = NULL;
	struct arrowkeys_list_t *new_node = NULL;

	struct arrowkeys_file_list_t *prev_file = NULL;

	while (!err && fgets(buf, sizeof(buf), fp) != NULL) {
		line++;
		char *tok = buf;

		while (*tok == ' ' || *tok == '\t' || *tok == '\r' || *tok == '\n') {
			tok++;
		}

		if (*tok == '#' || *tok == ';' || *tok == 0) {
			continue;
		}

		// R{{bNXx
		if (*tok == '[') {
			if (new_node != NULL) {
				prev = new_node;
			}
			new_node = NULL;
			prev_file = NULL;

			tok++;
			char *end = tok + strlen(tok);
			while (tok < end) {
				end--;
				if (*end != ' ' && *end != '\t' && *end != '\r' && *end != '\n') {
					break;
				}
			}
			if (*end != ']') {
				err = true;
				break;
			}
			*end = 0;

			if (*tok == 0) {
				err = true;
				break;
			}

			new_node = new struct arrowkeys_list_t;
			new_node->next = NULL;
			new_node->file = NULL;
			new_node->replacekeys[ARROWKEYS_INDEX_LEFT ].key = VK_LEFT;
			new_node->replacekeys[ARROWKEYS_INDEX_LEFT ].is_ascii = NOT_ASCII;
			new_node->replacekeys[ARROWKEYS_INDEX_UP   ].key = VK_UP;
			new_node->replacekeys[ARROWKEYS_INDEX_UP   ].is_ascii = NOT_ASCII;
			new_node->replacekeys[ARROWKEYS_INDEX_RIGHT].key = VK_RIGHT;
			new_node->replacekeys[ARROWKEYS_INDEX_RIGHT].is_ascii = NOT_ASCII;
			new_node->replacekeys[ARROWKEYS_INDEX_DOWN ].key = VK_DOWN;
			new_node->replacekeys[ARROWKEYS_INDEX_DOWN ].is_ascii = NOT_ASCII;
			new_node->label = new char[end - tok + 1];
			strcpy(new_node->label, tok);

			if (list == NULL) {
				list = new_node;
			} else {
				prev->next = new_node;
			}
			size++;

			continue;
		}

		tok = strtok(tok, " \t\r\n");
		if (tok == NULL) {
			continue;
		}

		if (new_node == NULL) {
			err = true;
			break;
		}

		int index;
		bool is_file = false;
		bool is_dir = false;
		if (stricmp(tok, "left") == 0) {
			index = ARROWKEYS_INDEX_LEFT;
		} else if (stricmp(tok, "up") == 0) {
			index = ARROWKEYS_INDEX_UP;
		} else if (stricmp(tok, "right") == 0) {
			index = ARROWKEYS_INDEX_RIGHT;
		} else if (stricmp(tok, "down") == 0) {
			index = ARROWKEYS_INDEX_DOWN;
		} else if (stricmp(tok, "file") == 0) {
			is_file = true;
		} else if (stricmp(tok, "dir") == 0) {
			is_dir = true;
		} else {
			// sL[[h
			err = true;
			break;
		}

		if (is_file || is_dir) {
			// t@C or fBNg
			// r̋󔒂ŋ؂Ȃ悤Ɏc̕ĎOőŐ󔒂
			tok = strtok(NULL, "\r\n");
			if (tok == NULL) {
				err = true;
				break;
			}

			while (*tok == ' ' || *tok == '\t') {
				tok++;
			}
			if (*tok == 0) {
				err = true;
				break;
			}

			char *end = tok + strlen(tok);
			while (tok < end-- && (*end == ' ' || *end == '\t')) {
				*end = 0;
			}
			if (*tok == 0) {
				err = true;
				break;
			}

			struct arrowkeys_file_list_t *new_file = new struct arrowkeys_file_list_t;
			if (new_node->file == NULL) {
				new_node->file = new_file;
			} else {
				prev_file->next = new_file;
			}
			prev_file = new_file;

			new_file->next = NULL;
			new_file->is_dir = is_dir;
			new_file->name = new char[strlen(tok) + 1];
			strcpy(new_file->name, tok);
		} else {
			// L[蓖
			tok = strtok(NULL, " \t\r\n");
			if (tok == NULL) {
				err = true;
				break;
			}

			if (0x20 < *tok && *tok < 0x7F && *(tok + 1) == 0) {
				new_node->replacekeys[index].key = *tok;
			} else if (stricmp(tok, "RETURN") == 0) {
				new_node->replacekeys[index].key = 0x0D;
			} else if (stricmp(tok, "SPACE") == 0) {
				new_node->replacekeys[index].key = ' ';
			} else if (stricmp(tok, "TAB") == 0) {
				new_node->replacekeys[index].key = '\t';
			} else {
				err = true;
				break;
			}
			new_node->replacekeys[index].is_ascii = ASCII;
		}
	}
	fclose(fp);

	if (err) {
		KeybFreeArrowKeysList(new_node);
		KeybFreeArrowKeysList(list);
		list = NULL;

		return -1;
	}

	return size;
}

//===========================================================================
LPTSTR KeybCreateArrowKeysChoices()
{
	return KeybCreateArrowKeysChoices(NULL);
}

//===========================================================================
LPTSTR KeybCreateArrowKeysChoices(const struct arrowkeys_list_t *list)
{
	if (list == NULL) {
		list = g_arrowkeys_list;
	}

	LPCTSTR label_no_changes = TEXT("no changes");
	LPCTSTR label_auto_select = TEXT("<auto select>");

	int len = _tcslen(label_no_changes) + 1 + _tcslen(label_auto_select) + 1;
	const struct arrowkeys_list_t *save = list;
	for ( ; list != NULL; list = list->next) {
		len += strlen(list->label) + 1;
	}
	len++;
	list = save;

	LPTSTR choices = new TCHAR[len];
	LPTSTR ptr = choices;

	_tcscpy(ptr, label_no_changes);
	ptr += _tcslen(label_no_changes) + 1;

	_tcscpy(ptr, label_auto_select);
	ptr += _tcslen(label_auto_select) + 1;

	for ( ; list != NULL; list = list->next) {
		_tcscpy(ptr, list->label);
		ptr += _tcslen(list->label) + 1;
	}
	*ptr = 0;

	return choices;
}

//===========================================================================
bool KeybArrowKeysIsAutoSelect()
{
	return (g_nArrowKeys == ARROWKEYS_AUTO_SELECT);
}

//===========================================================================
bool KeybSelectArrowKeys()
{
	if (g_nArrowKeys <= ARROWKEYS_NO_CHANGES) {
		// ύXȂ
		current_replacekeys = default_replacekeys;
		return true;
	} else if (g_nArrowKeys == ARROWKEYS_AUTO_SELECT) {
		// ݂̊蓖Ă̂܂
		return false;
	}

	int index = g_nArrowKeys - ARROWKEYS_INDEX_OFFSET;

	for (const struct arrowkeys_list_t *list = g_arrowkeys_list; list != NULL; list = list->next) {
		if (index-- == 0) {
			current_replacekeys = list->replacekeys;
			return true;
		}
	}

	// ύXȂ
	current_replacekeys = default_replacekeys;
	return false;
}

//===========================================================================
bool KeybSelectArrowKeys(LPCTSTR filename)
{
	if (filename == NULL || *filename == 0) {
		// ύXȂ
		current_replacekeys = default_replacekeys;
		return false;
	}

	LPCTSTR dir = filename;
	int dir_len;
	LPCTSTR file = _tcschr(dir, _T('\\'));
	if (file == NULL) {
		dir = NULL;
		dir_len = 0;
		file = filename;
	} else {
		dir_len = file - dir;
		file++;
		LPCTSTR next;
		while ((next = _tcschr(file, _T('\\'))) != NULL) {
			dir = file;
			dir_len = next - dir;
			file = ++next;
		}
	}

	// t@CɈvݒT
	for (const struct arrowkeys_list_t *list = g_arrowkeys_list; list != NULL; list = list->next) {
		for (const struct arrowkeys_file_list_t *file_list = list->file; file_list != NULL; file_list = file_list->next) {
			if (file_list->is_dir) {
				// fBNg
				if (dir != NULL && _tcsnicmp(file_list->name, dir, dir_len) == 0 && file_list->name[dir_len] == 0) {
					current_replacekeys = list->replacekeys;
					return true;
				}
			} else {
				// t@C(Ov)
				if (_tcsnicmp(file_list->name, file, _tcslen(file_list->name)) == 0) {
					current_replacekeys = list->replacekeys;
					return true;
				}
			}
		}
	}

	// ύXȂ
	current_replacekeys = default_replacekeys;
	return false;
}

#endif	// +CU version
